-- M.A.I.D.  (Mining And Inventory Daemon)
--
-- Manage ship cargo, ores and mining.
--
-- Features:
--   Automatically jettison items of your choosing (ores, cargo, scrap, etc).
--   Jettison items immediately (combat mode) or just before ship is full.
--   Time cargo drops to make sure they don't disappear.
--   Monitor asteroid temperatures to prevent overheating.
--   Review mining and cargo statistics.
--
--
-- Binds:
--   /maid                Open M.A.I.D. window
--   /maid.toggle         Toggle the automatic cargo jettison
--
--
-- Known bugs (just nuisances, really):
--   MAID can't tell the difference between ore that was mined and ore that
--   was collected.  If you pick up a single crate of ore, MAID will think you
--   mined it.  Thus, the cargo tally may show that there is ore you haven't
--   picked up, even if you did, and the cargo timer won't automatically stop.
--
--
-- Assembled and written by igrok
-- AutoJett originally by slime73
-- Portions adopted from AutoJett v2.3 GUI by iry
-- Cargo timer adopted from Timer by Chocolateer
--
--

MAID = {}
MAID.version = '1.3'

MAID.settings = {}


-- Automatically jettison cargo?
MAID.settings.jettison = true

-- Exclusive mode:
-- enabled      Assume that all items not on the cargo list should be jettisoned.
-- disabled     Only jettison the items marked in the cargo list.
MAID.settings.exclusiveMode = false

-- Combat jettison mode:
-- enabled      Jettison anything and everything immediately.
-- disabled     Only jettison those items selected by the user.
MAID.settings.combatMode = false

-- Notify user when asteroid has overheated?
MAID.settings.tempNotify = true

-- Automatically stop mining to prevent asteroid overheating?
MAID.settings.tempSafety = false

-- Maximum temperature of asteroid (K)
MAID.settings.tempMax = 100

-- Play sound notifications?
MAID.settings.sound = true

-- Print debug statements?
MAID.settings.debug = false


-- load sound effects
MAID.sounds = {
    ['burning'] = false,
    ['timer_end'] = false,
    ['ship_full'] = false,
    ['jettison'] = false,
}
for sound,loaded in pairs(MAID.sounds) do
    local fileStr = 'plugins/MAID/sounds/' .. sound .. '.ogg'

    local result = gksound.GKLoadSound{soundname = sound, filename = fileStr}

    if result then
        MAID.sounds[sound] = true
    end
end


-- Table of items to keep or jettison.
MAID.cargoItems = {}

-- Keep track of cargo currently in ship.
-- 'Name' -> Quantity
MAID.currentCargo = {}

-- Keep track of cargo that's recently been in the hold.
-- 'Name' -> 1
MAID.recentCargo = {}

-- Keep track of ore that's recently been mined - to keep the drop counts correct.
-- We can't tell the difference between ore that's been picked up and
-- ore that's been mined, so we have to track it manually...
-- {'Name' -> 1}
MAID.minedOre = {}



function MAID:print(message)
    if message then
        HUD:PrintSecondaryMsg('\12700ffaa' .. message ..'\127o')
    end
end
function MAID:debugPrint(message)
    if MAID.settings.debug and message then
        purchaseprint('\127ff0000' .. message ..'\127o')
    end
end


dofile("settings.lua")
dofile("time.lua")
dofile("ui.lua")
dofile("ui.addItemPopup.lua")


-- keep track of cargo drops
function MAID:registerCargoDrop(itemname, quantity)
    for i,item in ipairs(MAID.cargoItems) do
        if item.Name == itemname then
            MAID.cargoItems[i]['Dropped'] = quantity + MAID.cargoItems[i]['Dropped']
            MAID.time:startTimer(itemname)

            -- When items are collected and then dropped again, the cargo collection/drop
            -- numbers are all out of skew.  We can reconcile most (all?) numbers.
            if MAID.cargoItems[i]['Collected'] >= quantity then
                MAID.cargoItems[i]['Collected'] = MAID.cargoItems[i]['Collected'] - quantity
            else
                MAID.cargoItems[i]['Collected'] = 0
            end

            MAID:debugPrint('MAID: ' .. quantity .. ' units of ' .. itemname .. ' were dropped.')
            MAID.UI:reset_data()
            return
        end
    end
end


-- keep track of cargo collections
function MAID:registerCargoCollection(itemname, quantity)
    for i,item in ipairs(MAID.cargoItems) do
        if item.Name == itemname then
            MAID.cargoItems[i]['Collected'] = quantity + MAID.cargoItems[i]['Collected']
            MAID:debugPrint('MAID: ' .. quantity .. ' units of ' .. itemname .. ' were collected.')

            -- When items are dropped and then collected again, the cargo collection/drop
            -- numbers are all out of skew.  We can reconcile most (all?) numbers.
            -- This is further complicated by the fact that ores can be mined or picked up.
            if quantity == 1 and itemname:match(' Ore') and MAID.minedOre[itemname] ~= nil and MAID.minedOre[itemname] > 0 then
                MAID:debugPrint('MAID: ' .. MAID.minedOre[itemname] .. ' units of ' .. itemname .. ' have been mined.')

                -- If a crate of ore has been mined, but some has already been dropped, we don't want to decrement the dropped number.
                if MAID.cargoItems[i]['Dropped'] > 0 then
                    MAID.minedOre[itemname] = MAID.minedOre[itemname] - 1
                end

            -- Everything but a single crate of ore will fall through to here.
            -- If the item is ore, but is greater than quantity 1, it was not mined
            elseif MAID.cargoItems[i]['Dropped'] >= quantity then
                MAID.cargoItems[i]['Dropped'] = MAID.cargoItems[i]['Dropped'] - quantity
            elseif MAID.cargoItems[i]['Dropped'] > 0 then
                MAID.cargoItems[i]['Dropped'] = 0
            end

            -- don't kill the cargo timer unless all the cargo has been picked up
            if MAID.cargoItems[i]['Dropped'] == 0 then
                MAID.time:stopTimer(itemname)
            end

            MAID.UI:reset_data()
            return
        end
    end

    -- if the item is not in the list of tracked cargo, we'll save its name
    MAID.recentCargo[itemname] = 1
    MAID:debugPrint('MAID: added ' .. itemname .. ' to the list of recent cargo.')
end


-- initialize the cargo counts, given the current cargo in the ship's hold
function MAID:initializeCurrentCargo(cargoids)
    MAID:debugPrint('MAID initializing current cargo manifest.')

    MAID.currentCargo = {}

    -- go through each piece of cargo currently in the hold
    for i,itemid in ipairs(cargoids) do
        local itemname = GetInventoryItemName(itemid)
        local qty = GetInventoryItemQuantity(itemid)
        MAID.currentCargo[itemname] = qty
    end
end


-- Check to see if any of the current cargo was mined, rather than picked up.
-- This function complements MAID:updateCurrentCargo(), so it does not increment
-- the cargo counters.  It merely determines (as best it can) how much was mined.
function MAID:checkMinedCargo(itemname, quantity)
    -- It's best to check if the last item added was a single crate of ore.
    -- If so we have to assume it was mined.
    if quantity == 1 then
        if itemname:match(' Ore') then
            if MAID.minedOre[itemname] == nil then
                MAID.minedOre[itemname] = 1
            else
                MAID.minedOre[itemname] = MAID.minedOre[itemname] + 1
            end
        end
    end
end


-- update the cargo counts, given the current cargo in the ship's hold
function MAID:updateCurrentCargo(cargoids)
    local previousCargo = MAID.currentCargo

    -- build index of new cargo; will become MAID.currentCargo
    local newCargo = {}

    -- go through each piece of cargo currently in the hold
    for i,itemid in ipairs(cargoids) do
        local itemname = GetInventoryItemName(itemid)
        local qty = GetInventoryItemQuantity(itemid)
        newCargo[itemname] = qty

        if previousCargo[itemname] ~= nil then
            if previousCargo[itemname] < newCargo[itemname] then
                local collected = newCargo[itemname] - previousCargo[itemname]
                MAID:checkMinedCargo(itemname, collected)
                MAID:registerCargoCollection(itemname, collected)
            elseif previousCargo[itemname] > newCargo[itemname] then
                local dropped = previousCargo[itemname] - newCargo[itemname]
                MAID:registerCargoDrop(itemname, dropped)
            end
        else
            MAID:checkMinedCargo(itemname, qty)
            MAID:registerCargoCollection(itemname, qty)
        end
    end

    -- go through each piece of cargo that used to be in the hold
    -- this is to catch items that are no longer in the current hold
    for itemname,qty in pairs(previousCargo) do
        if newCargo[itemname] == nil then
            MAID:registerCargoDrop(itemname, qty)
        end
    end

    MAID.currentCargo = newCargo
end


-- Get the weapon group and trigger number for an add-on.
function MAID:GetWeapongroupAndTriggerNumberFromGroupID(gid)
    if 0<=gid and gid<=5 then
        return gid+1,2
    elseif 6<=gid and gid<=11 then
        return gid-5, 1
    elseif 12<=gid and gid<=17 then
        return gid-11, 3
    end
end


-- Disable the mining beam at the specified port.
function MAID:disableBeamAtPort(portid)
MAID.ActiveShipWeaponGroup = {}
    for gid=0,17 do
        for k,v in pairs(GetActiveShipWeaponGroup(gid)) do
            if k==portid then
                local weapongroup, trigger = MAID:GetWeapongroupAndTriggerNumberFromGroupID(gid)

                -- the below commands are the equivalent of setting the proper group,
                -- pressing the fire button and releasing the fire button
                gkinterface.GKProcessCommand('weapon'..weapongroup)
                gkinterface.GKProcessCommand('+Shoot'..trigger..' 1')
                -- For some reason, we can't simulate the key release!?
                --gkinterface.GKProcessCommand('+Shoot'..trigger..' 0')
                return
            end
        end
    end
end


function MAID:OnEvent(event, data)
    if (event == "PLAYER_ENTERED_GAME" and not PlayerInStation()) or event == "LEAVING_STATION" then
        -- when entering game or leaving station, initialize cargo based on what's already in the hold
        local shipinventory = GetShipInventory(GetActiveShipID())
        local cargoids = shipinventory.cargo
        MAID:initializeCurrentCargo(cargoids)
    elseif event == "INVENTORY_REMOVE" and not PlayerInStation() then
        -- see if anything has changed in the inventory
        local shipinventory = GetShipInventory(GetActiveShipID())
        local cargoids = shipinventory.cargo
        MAID:updateCurrentCargo(cargoids)
    -- monitor cargo inventory and jettison as needed
    elseif (event == "INVENTORY_ADD" or event == "INVENTORY_UPDATE") and not PlayerInStation() then
        -- see if anything has changed in the inventory
        local shipinventory = GetShipInventory(GetActiveShipID())
        local cargoids = shipinventory.cargo
        MAID:updateCurrentCargo(cargoids)

        if MAID.settings.jettison and next(cargoids) ~= nil then
            -- if we're in combat mode, jettison everything now
            if MAID.settings.combatMode then
                JettisonAll()

                if MAID.settings.sound and MAID.sounds['jettison'] then
                    gksound.GKPlaySound('jettison')
                end
            else
                -- Play a notification if the ship is full of cargo.
                local currentCargo = GetActiveShipCargoCount()
                local maxCargo = GetActiveShipMaxCargo()
                if maxCargo == currentCargo and MAID.settings.sound and MAID.sounds['ship_full'] then
                    gksound.GKPlaySound('ship_full')
                    MAID:print('M.A.I.D: cargo hold is full.')
                end


                -- to prevent huge numbers of single crates, do not dump cargo right away

                -- for some reason, the ship and cargo are not properly initialized on login
                -- so we will get to this point, even if the ship is in a station
                if maxCargo == nil or currentCargo == nil then return end

                local availableSpace = 0
                availableSpace = maxCargo - currentCargo

                if currentCargo > 0 and availableSpace <= 5 then
                    -- go through each piece of cargo in the hold
                    for i,itemid in ipairs(cargoids) do
                        local itemname = GetInventoryItemName(itemid)

                        -- go through each type of cargo being tracked
                        local found = false
                        for n,cargotype in ipairs(MAID.cargoItems) do
                            -- if item is in cargo jettison list, and marked, we jettison
                            if cargotype['Name'] == itemname and cargotype['Jettison'] then
                                local qty = GetInventoryItemQuantity(itemid)
                                JettisonMultiple{{id=itemid, quantity=qty}}

                                if MAID.settings.sound and MAID.sounds['jettison'] then
                                    gksound.GKPlaySound('jettison')
                                end

                            -- if in exclusive mode, we need to check if the item is in the cargo list
                            elseif MAID.settings.exclusiveMode and cargotype['Name'] == itemname then
                                found = true
                            end
                        end

                        -- if in exclusive mode, and the item is not in the cargo list, we jettison
                        if MAID.settings.exclusiveMode and not found then
                            local qty = GetInventoryItemQuantity(itemid)
                            JettisonMultiple{{id=itemid, quantity=qty}}

                            if MAID.settings.sound and MAID.sounds['jettison'] then
                                gksound.GKPlaySound('jettison')
                            end

                            MAID:debugPrint('MAID (exclusive mode): ' .. qty .. ' units of ' .. itemname .. ' were dropped.')
                        end
                    end
                end
            end
        end

    -- monitor asteroid for overheating during mining
    elseif event == "TARGET_SCANNED" then
        if MAID.settings.tempNotify or MAID.settings.tempSafety then        
            if table.maxn(HUD.progressbars.items) > 0 then
                local roidtemp = tonumber(string.match(data, 'Temperature: (%d+%.%d+) K'))
                if roidtemp > MAID.settings.tempMax then
                    -- if safety is on, disable mining beams
                    if MAID.settings.tempSafety then
                        MAID:print('M.A.I.D: Disabling mining beams - asteroid temperature at ' .. roidtemp .. 'K.')
                        for k,v in pairs(HUD.progressbars.items) do 
                            MAID:disableBeamAtPort(GetActiveShipPortIDOfItem(k))
                        end

                    -- if notifications are on, print a warning
                    elseif MAID.settings.tempNotify then 
                        MAID:print('M.A.I.D: Warning! Asteroid is overheated - temperature at ' .. roidtemp .. 'K.')
                    end

                    -- if sound is on, play a warning
                    if MAID.settings.sound and MAID.sounds['burning'] then
                        gksound.GKPlaySound("burning")
                    end
                end
            end
        end
    end
end


-- Toggle jettison on and off
function MAID:Toggle()
    if MAID.settings.jettison then
        MAID.settings.jettison = false
        MAID:print("M.A.I.D: jettison disabled")
    else
        MAID.settings.jettison = true
        MAID:print("M.A.I.D: jettison enabled")
    end
end

-- Toggle exclusive mode on and off
function MAID:ToggleExclusiveMode()
    if MAID.settings.exclusiveMode then
        MAID.settings.exclusiveMode = false
        MAID:print("M.A.I.D: exclusive mode disabled")
    else
        MAID.settings.exclusiveMode = true
        MAID:print("M.A.I.D: exclusive mode enabled")
    end
end

-- Toggle combat mode on and off
function MAID:ToggleCombatMode()
    if MAID.settings.combatMode then
        MAID.settings.combatMode = false
        MAID:print("M.A.I.D: combat mode disabled")
    else
        JettisonAll()
        if MAID.settings.sound and MAID.sounds['jettison'] then
            gksound.GKPlaySound('jettison')
        end

        MAID.settings.combatMode = true
        MAID:print("M.A.I.D: combat mode enabled")
    end
end

-- Toggle temperature notifications on and off
function MAID:ToggleTempNotify()
    if MAID.settings.tempNotify then
        MAID.settings.tempNotify = false
        MAID:print("M.A.I.D: temperature notifications disabled")
    else
        MAID.settings.tempNotify = true
        MAID:print("M.A.I.D: temperature notifications enabled")
    end
end

-- Toggle temperature safety on and off
function MAID:ToggleTempSafety()
    if MAID.settings.tempSafety then
        MAID.settings.tempSafety = false
        MAID:print("M.A.I.D: temperature safety disabled")
    else
        MAID.settings.tempSafety = true
        MAID:print("M.A.I.D: temperature safety enabled")
    end
end

-- Set safe asteroid temperature
function MAID:SetSafeTemperature(value)
    local temperature = tonumber(value)
    if temperature and type(temperature) == 'number' then
        MAID.settings.tempMax = temperature
        MAID:print("M.A.I.D: safe temperature set to: " .. tostring(MAID.settings.tempMax) .. "K.")
    end
end

-- Toggle sound on and off
function MAID:ToggleSound()
    if MAID.settings.sound then
        MAID.settings.sound = false
        MAID:print("M.A.I.D: sound disabled")
    else
        MAID.settings.sound = true
        MAID:print("M.A.I.D: sound enabled")
    end
end

-- Toggle debug on and off
function MAID:ToggleDebug()
    if MAID.settings.debug then
        MAID.settings.debug = false
        MAID:print("M.A.I.D: debug disabled")
    else
        MAID.settings.debug = true
        MAID:print("M.A.I.D: debug enabled")
    end
end


RegisterEvent(MAID, "PLAYER_ENTERED_GAME")
RegisterEvent(MAID, "LEAVING_STATION")
RegisterEvent(MAID, "INVENTORY_REMOVE")
RegisterEvent(MAID, "INVENTORY_ADD")
RegisterEvent(MAID, "INVENTORY_UPDATE")
RegisterEvent(MAID, "TARGET_SCANNED")

RegisterUserCommand('maid', function()
        if MAID.settings.initializing then
            MAID.settings:read()
            MAID.settings.initializing = false

            -- once settings have been loaded, we don't need these events
            UnregisterEvent(MAID.settings, "SECTOR_LOADED")
        end
        return MAID.UI:Show()
    end)
RegisterUserCommand('maid.toggle', function() return MAID:Toggle() end)
RegisterUserCommand('maid.toggleExclusiveMode', function() return MAID:ToggleExclusiveMode() end)
RegisterUserCommand('maid.toggleCombatMode', function() return MAID:ToggleCombatMode() end)
RegisterUserCommand('maid.toggleTempNotify', function() return MAID:ToggleTempNotify() end)
RegisterUserCommand('maid.toggleTempSafety', function() return MAID:ToggleTempSafety() end)
RegisterUserCommand('maid.toggleSound', function() return MAID:ToggleSound() end)
RegisterUserCommand('maid.toggleDebug', function() return MAID:ToggleDebug() end)

